#include <Arduino.h>
#include <Wire.h>
#include <MPU6050.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <SPI.h>
#include <LiquidCrystal_I2C.h>

bool exploded;

const long totalTime = 60000;  //180000;
unsigned long startTime;
unsigned long currentTime;
unsigned long elapsedTime;

String series = "";

#define LCD_ADDRESS 0x3F
LiquidCrystal_I2C lcd(LCD_ADDRESS, 16, 2);


// SPI DISPLAY
#define SCREEN_WIDTH 128 
#define SCREEN_HEIGHT 64

#define OLED_CLK 18
#define OLED_MOSI 23
#define OLED_CS 5
#define OLED_DC 1
#define OLED_RESET 3
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT,
                         OLED_MOSI, OLED_CLK, OLED_DC, OLED_RESET, OLED_CS);

const int mazeWidth = 16;
const int mazeHeight = 8;

const int numMazes = 4;
int8_t mazes[numMazes][mazeHeight][mazeWidth] = {
  { // Maze 1
    { 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
    { 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1 },
    { 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 0, 1 },
    { 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1 },
    { 1, 0, 1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 1, 1 },
    { 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1 },
    { 1, 0, 1, 1, 1, 1, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1 },
    { 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } },
  { // Maze 2
    { 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
    { 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
    { 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1 },
    { 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1 },
    { 1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 1 },
    { 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1 },
    { 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1 },
    { 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 } },
  { // Maze 3
    { 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1 },
    { 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 0, 0, 0, 0, 1 },
    { 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1 },
    { 1, 0, 1, 1, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0 },
    { 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0 },
    { 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0 },
    { 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0 } },
  { // Maze 4
    { 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
    { 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1 },
    { 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 1 },
    { 1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 1, 0, 1 },
    { 1, 0, 1, 0, 0, 0, 1, 0, 1, 1, 0, 1, 1, 1, 0, 1 },
    { 1, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 1, 0, 1 },
    { 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1 },
    { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0 } }
};

int upButtonPin = 12;
int downButtonPin = 15;
int rightButtonPin = 13;
bool mazeFinished = false;

int playerX = 0;
int playerY = 0;
int mazeIndex;

void displaySetup() {
  if (!display.begin(SSD1306_SWITCHCAPVCC)) {
    Serial.println(F("SSD1306 allocation failed"));
    for (;;)
      ; 
  }

  display.clearDisplay();
  randomSeed(analogRead(0)); 

  switch ((series[3] - '0') % 3 == 0) {
    case 1:
      upButtonPin = 13;
      downButtonPin = 12;
      rightButtonPin = 15;
      break;
    case 2:
      upButtonPin = 15;
      downButtonPin = 13;
      rightButtonPin = 12;
      break;
    default:
      break;
  }

  pinMode(upButtonPin, INPUT_PULLUP);
  pinMode(downButtonPin, INPUT_PULLUP);
  pinMode(rightButtonPin, INPUT_PULLUP);

  mazeIndex = random(numMazes); 
  drawMaze(mazeIndex);
}

unsigned long lastDebounceTimesPlayer[3] = {0, 0, 0};
const unsigned long debounceDelayPlayer = 50;

void updatePlayerPosition() {
  static int previousButtonStates[3] = {HIGH, HIGH, HIGH};
  int currentButtonStates[3] = {
    digitalRead(upButtonPin),
    digitalRead(downButtonPin),
    digitalRead(rightButtonPin),
  };

  if (currentButtonStates[0] != previousButtonStates[0] && (millis() - lastDebounceTimesPlayer[0]) > debounceDelayPlayer) {
    if (currentButtonStates[0] == LOW && playerY > 0 && mazes[mazeIndex][playerY - 1][playerX] == 0) {
      playerY--;
    } else if (currentButtonStates[0] == LOW) {
      exploded = 1;
    }
    lastDebounceTimesPlayer[0] = millis();
  }

  if (currentButtonStates[1] != previousButtonStates[1] && (millis() - lastDebounceTimesPlayer[1]) > debounceDelayPlayer) {
    if (currentButtonStates[1] == LOW && playerY < mazeHeight - 1 && mazes[mazeIndex][playerY + 1][playerX] == 0) {
      playerY++;
    } else if (currentButtonStates[1] == LOW) {
      exploded = 1;
    }
    lastDebounceTimesPlayer[1] = millis();
  }

  if (currentButtonStates[2] != previousButtonStates[2] && (millis() - lastDebounceTimesPlayer[2]) > debounceDelayPlayer) {
    if (currentButtonStates[2] == LOW && playerX < mazeWidth - 1 && mazes[mazeIndex][playerY][playerX + 1] == 0) {
      playerX++;
    } else if (currentButtonStates[1] == LOW) {
      exploded = 1;
    }
    lastDebounceTimesPlayer[2] = millis();

    if (playerX == mazeWidth - 1 && playerY == mazeHeight - 1) {
      mazeFinished = 1;
    }
  }


  for (int i = 0; i < 3; i++) {
    previousButtonStates[i] = currentButtonStates[i];
  }

  drawMaze(mazeIndex);
}

void labyrinthTick() {
  updatePlayerPosition();
}

void drawMaze(int mazeIndex) {
  display.clearDisplay();
  int cellWidth = SCREEN_WIDTH / mazeWidth;
  int cellHeight = SCREEN_HEIGHT / mazeHeight;

  for (int y = 0; y < mazeHeight; ++y) {
    for (int x = 0; x < mazeWidth; ++x) {
      if (mazes[mazeIndex][y][x] == 1) {
        display.fillRect(x * cellWidth, y * cellHeight, cellWidth, cellHeight, SSD1306_WHITE);
      } else {
        display.fillRect(x * cellWidth, y * cellHeight, cellWidth, cellHeight, SSD1306_BLACK);
      }
    }
  }

  int playerXPos = playerX * cellWidth + cellWidth / 2;
  int playerYPos = playerY * cellHeight + cellHeight / 2;
  int radius = min(cellWidth, cellHeight) / 2 - 2;
  display.drawCircle(playerXPos, playerYPos, radius, SSD1306_WHITE);

  int endX = (mazeWidth - 1) * cellWidth;
  int endY = (mazeHeight - 1) * cellHeight;
  display.drawLine(endX, endY, endX + cellWidth, endY + cellHeight, SSD1306_WHITE);
  display.drawLine(endX + cellWidth, endY, endX, endY + cellHeight, SSD1306_WHITE);

  display.display();
}

// BUZZER
const int buzzerPin = 27;
unsigned long previousBuzzerTick = 0;
long beepInterval = 1000;

void buzzerSetup() {
  previousBuzzerTick = startTime;
  pinMode(buzzerPin, OUTPUT);
}

void bombTick() {
  // tone(buzzerPin, 1000, 50);
}

void explosionSound() {
  for (int i = 100; i < 2000; i += 100) {
    tone(buzzerPin, i, 50);
    delay(50);
  }
  for (int i = 2000; i > 100; i -= 100) {
    tone(buzzerPin, i, 50);
    delay(50);
  }
  noTone(buzzerPin);
}

void playGameWinningSound() {
  const int melody[] = {
    262, 294, 330, 349, 392, 440, 494, 523
  };

  const int noteDurations[] = {
    4, 4, 4, 4, 4, 4, 4, 4
  };

  const int melodyLength = sizeof(melody) / sizeof(melody[0]);

  for (int thisNote = 0; thisNote < melodyLength; thisNote++) {
    int noteDuration = 1000 / noteDurations[thisNote];
    tone(buzzerPin, melody[thisNote], noteDuration);

    int pauseBetweenNotes = noteDuration * 1.30;
    delay(pauseBetweenNotes);

    noTone(buzzerPin);
  }
}

void timeTick() {
  long remainingTime = totalTime - elapsedTime;

  if (remainingTime <= 0) {
    Serial.println("Bomb has gone off!");
    exploded = true;
    return;
  }

  beepInterval = 100 + (900 * remainingTime) / totalTime;

  if (currentTime - previousBuzzerTick >= beepInterval) {
    previousBuzzerTick = currentTime;
    bombTick();
  }

  lcd.setCursor(0, 1);
  lcd.print("Time: ");
  lcd.print(remainingTime / 1000);
  lcd.print(" s    "); 
}

// GYROSCOPE & I2C DISPLAY
MPU6050 mpu;

const int16_t accelThreshold = 9000;
const int16_t gyroThreshold = 9000;

int16_t prevAx, prevAy, prevAz;
int16_t prevGx, prevGy, prevGz;

void generateSeries() {
  randomSeed(analogRead(0));
  series = "";

  char firstByte = "ABCDEF"[random(0, 6)];
  char secondByte = "ABCDEF"[random(0, 6)];

  char thirdByte = "0123456789"[random(0, 10)];
  char fourthByte = "0123456789"[random(0, 10)];

  series += firstByte;
  series += secondByte;
  series += thirdByte;
  series += fourthByte;
}

void takeInitialReadings() {
  mpu.getMotion6(&prevAx, &prevAy, &prevAz, &prevGx, &prevGy, &prevGz);
  Serial.println("Initial readings taken:");
}

void gyroSetup() {
  Wire.begin(21, 22);

  lcd.init();
  lcd.backlight();

  Serial.println("Initializing the MPU6050...");
  mpu.initialize();

  if (mpu.testConnection()) {
    Serial.println("MPU6050 connection successful");
  } else {
    Serial.println("MPU6050 connection failed");
    while (1)
      ;
  }

  delay(50);
  takeInitialReadings();
  generateSeries();
  lcd.setCursor(0, 0);
  lcd.print("Series: ");
  lcd.print(series);
  lcd.setCursor(0, 1);
}

void gyroReadings() {
  int16_t ax, ay, az, gx, gy, gz;
  mpu.getMotion6(&ax, &ay, &az, &gx, &gy, &gz);

  int16_t diffAx = abs(ax - prevAx);
  int16_t diffAy = abs(ay - prevAy);
  int16_t diffAz = abs(az - prevAz);
  int16_t diffGx = abs(gx - prevGx);
  int16_t diffGy = abs(gy - prevGy);
  int16_t diffGz = abs(gz - prevGz);

  if (diffAx > accelThreshold || diffAy > accelThreshold || diffAz > accelThreshold || diffGx > gyroThreshold || diffGy > gyroThreshold || diffGz > gyroThreshold) {
    Serial.println("Wild movement detected!");
    Serial.print("diffAx: ");
    Serial.print(diffAx);
    Serial.print(", diffAy: ");
    Serial.print(diffAy);
    Serial.print(", diffAz: ");
    Serial.print(diffAz);
    Serial.print(", diffGx: ");
    Serial.print(diffGx);
    Serial.print(", diffGy: ");
    Serial.print(diffGy);
    Serial.print(", diffGz: ");
    Serial.println(diffGz);
    exploded = true;
  } else {
    // Serial.println("No significant movement.");
  }

  int16_t tempRaw = mpu.getTemperature();
  float tempC = (tempRaw / 340.0) + 36.53;

  prevAx = ax;
  prevAy = ay;
  prevAz = az;
  prevGx = gx;
  prevGy = gy;
  prevGz = gz;
}

// LEDGAME
#define RED 0
#define GREEN 1
#define BLUE 2
#define YELLOW 3

const int redButtonPin = 14;
const int greenButtonPin = 0;
const int blueButtonPin = 4;
const int yellowButtonPin = 2;

const int redLEDPin = 32;
const int greenLEDPin = 33;
const int blueLEDPin = 26;
const int yellowLEDPin = 25;

int buttonPins[] = { redButtonPin, greenButtonPin, blueButtonPin, yellowButtonPin };
int ledPins[] = { redLEDPin, greenLEDPin, blueLEDPin, yellowLEDPin };

bool buttonStates[4] = { HIGH, HIGH, HIGH, HIGH };

unsigned long lastDebounceTimes[4] = { 0, 0, 0, 0 };
const unsigned long debounceDelay = 50;

const int sequence[] = { 0, 1, 2, 3 };
const int sequenceLength = sizeof(sequence) / sizeof(sequence[0]);

int buttonStep = 0;
int currentStep = 0;
int flashStep = 0;

bool ledGameEnd = false;
unsigned long lastFlashTime = 0;
unsigned long flashInterval = 2000;
unsigned long lastIndicateTime = 0;
unsigned long indicateInterval = 200;
int indicateStep = 0;

void shuffleArrays(int arr1[], int arr2[], int size) {
  for (int i = size - 1; i > 0; i--) {
    int j = random(0, i + 1);

    int temp = arr1[i];
    arr1[i] = arr1[j];
    arr1[j] = temp;

    temp = arr2[i];
    arr2[i] = arr2[j];
    arr2[j] = temp;
  }
}

void ledGameSetup() {
  ledGameEnd = false;
  shuffleArrays(buttonPins, ledPins, 4);

  for (int i = 0; i < 4; i++) {
    pinMode(ledPins[i], OUTPUT);
    digitalWrite(ledPins[i], LOW);
  }

  for (int i = 0; i < 4; i++) {
    pinMode(buttonPins[i], INPUT_PULLUP);
  }

  lastIndicateTime = millis();
}

void ledGameTick() {
  if (ledGameEnd)
    return;
  flashAllLEDsInOrder();
  checkPlayerInput();
}

void flashAllLEDsInOrder() {
  if (buttonStep == 0) {
    if (millis() - lastFlashTime >= flashInterval) {
      lastFlashTime = millis();
      digitalWrite(ledPins[flashStep], LOW);
      flashStep += 1;
      if (flashStep > currentStep)
        flashStep = 0;
    }

    if (millis() - lastFlashTime >= flashInterval / 2) {
      digitalWrite(ledPins[flashStep], LOW);
    } else {
      digitalWrite(ledPins[flashStep], HIGH);
    }
  }
}

void checkPlayerInput() {
  for (int i = 0; i < 4; i++) {
    int reading = digitalRead(buttonPins[i]);
    if (reading != buttonStates[i] && (millis() - lastDebounceTimes[i]) > debounceDelay) {
      if (buttonStates[i] == HIGH && reading == LOW) {
        Serial.println(buttonStep);
        Serial.println(i);
        Serial.println(sequence[buttonStep]);
        if (i == sequence[buttonStep]) {
          digitalWrite(ledPins[i], HIGH);
          delay(200);
          digitalWrite(ledPins[i], LOW);
          buttonStep++;
          if (buttonStep > currentStep) {
            ++currentStep;
            buttonStep = 0;
            flashStep = 0;
            lastFlashTime = millis();
          }
          if (currentStep == sequenceLength) {
            ledGameEnd = true;
            for (int i = 0; i < 4; ++i) {
              digitalWrite(ledPins[i], HIGH);
            }
          }
        } else {
          exploded = true;
          return;
        }
      }

      buttonStates[i] = reading;
      lastDebounceTimes[i] = millis();
    }
  }
}

void lose() {
  for (int i = 0; i < 4; i++) {
    digitalWrite(ledPins[i], LOW);
  }
  display.clearDisplay();
  display.display();

  // Clear the LCD display
  lcd.clear();
  explosionSound();
  while (1)
    ;
}

void displayOkSign() {
  display.clearDisplay();

  int centerX = SCREEN_WIDTH / 2;
  int centerY = SCREEN_HEIGHT / 2;
  int radius = 20;
  display.drawCircle(centerX, centerY, radius, SSD1306_WHITE);

  display.drawLine(centerX - 10, centerY, centerX - 2, centerY + 8, SSD1306_WHITE);
  display.drawLine(centerX - 2, centerY + 8, centerX + 10, centerY - 8, SSD1306_WHITE);

  display.display();
}

void win() {
  displayOkSign();
  playGameWinningSound();
  while (1)
    ;
}

void setup() {
  Serial.begin(115200);

  exploded = false;
  startTime = millis();

  displaySetup();
  buzzerSetup();
  gyroSetup();
  ledGameSetup();
}

void loop() {
  if (exploded) {
    lose();
  }
  if (ledGameEnd && mazeFinished) {
    win();
  }
  currentTime = millis();
  elapsedTime = currentTime - startTime;
  gyroReadings();
  timeTick();

  if (!ledGameEnd)
    ledGameTick();
  
  if (!mazeFinished)
    updatePlayerPosition();
}
